#!/usr/bin/perl -wT
##########
#
# samples.cgi Version 1.5
#
# Copyright 2004 Andreas Brinkmann
#                anbr@sci.kun.nl
#            
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
##########

use strict;

# You need the cgi-lib.pl library, put the correct path if necessary

require "cgi-lib.pl";

# Some global variables, modify for your local settings
#
# $TABLECOLS     Number of columns in the table. Attention, if you want to change
#                this change @options in LongMenu, and many other things!
# $INPUTCOLS     Number of columns in the text fields.
# $PATH          Path to the directory, containing the data and log files.
# $DATAFILE      Filename of the data file.
# $BACKUP        Filename of the backup file.
# $LOG           Extension of the log files.
# $LIMIT         Limit for the max length of text, especially for textareas!
# $ACTION        The path to this script on the web server.
# $CONTACT       The name of the person responsible for the database.
# $CONTACTMAIL   The mail address of the person responsible for the database.
# $REVISED       Latest date for updates to this script
# $BACKTO        Text for the link back to a main page
# $BACKTOLINK    Link back to a main page

my $TABLECOLS   = 9;
my $INPUTCOLS   = 60;
my $PATH        = "/users/anbr/Samples/";
my $DATAFILE    = "samples.dat";
my $BACKUP      = "samples.dat.bak";
my $LOG         = ".log";
my $LIMIT       = 1024;
my $ACTION      = "/~anbr/cgi-bin/test-samples.cgi";
my $CONTACT     = "Andreas";
my $CONTACTMAIL = "andreas\@solidmr.kun.nl";
my $REVISED     = "November 2002";
my $BACKTO      = "Back to the Software Page";
my $BACKTOLINK  = "/~anbr/software.html";

MAIN:
{
   my %in;
   my @data = ();   # data is an array which contains the data of the database.
                    # data[0] is the number of samples in the database
                    # data[1..] contains in serial the data, i.e.
                    #           data[1..$TABLECOLS]              first sample
                    #           data[$TABLECOLS+1..2*$TABLECOLS] second sample, etc.
   my $date;
   my $key;
   my $logfile;
   my @time;
   my $time;
   my $baktime;
   my $i;
	my $toomuchdata = 0;

   print &PrintHeader;
   print "<html>\n\n";
   print "<head>\n";
   print "   <title>Samples</title>\n";
   print '   <style type="text/css">',"\n";

   print "      body {\n";
   print "         margin-left: 2%;\n";
   print "         margin-right: 2%;\n";
   print	"         margin-top: 2%;\n";
   print "         margin-bottom: 2%;\n";
   print "         font-family: sans-serif;\n";
   print "         color: black;\n";
   print	"         background: #FFEEDD;\n";
   print "      }\n";

   print "      table.samples {\n";
   print "         font-family: sans-serif;\n";
   print "         background:   #FFEEDD;\n";
   print	"         border-color: #FFEEDD;\n";
   print "         font-size: 80%;\n";
   print "      }\n";

   print "      th.samples {\n";
   print "         font-family: sans-serif;\n";
   print "         background: #CCCC99;\n";
   print "         font-size: 80%;\n";
   print "      }\n";

   print "      td.samples {\n";
   print "         font-family: sans-serif;\n";
   print "         background: #F2E1D0;\n";
   print "         font-size: 80%;\n";
   print "      }\n";

   print "      p.redred {\n";
   print "         color : red;\n";
   print "         border: solid red;\n";
   print "      }\n";

   print "      div.redred {\n";
   print "         color : red;\n";
   print "         border: solid red;\n";
   print "      }\n";

   print "      div.redblack {\n";
   print "         color : black;\n";
   print "         border: solid green;\n";
   print "      }\n";

   print "   </style>\n";
   print "</head>\n\n";
   print '<body text="#000000" bgcolor="#FFEBCD" link="#0000EF"';
   print 'vlink="#55188A" alink="#FF 0000">',"\n\n";
   print "<center>\n\n";
   print "<h1>Samples</h1>\n\n";

   # Check if any parameters are passed to the program, the would appear in the
	# hash %in. Notice that the reference to the hash is passed to the function
	# ReadParse, i.e. \%in
   
   if (&ReadParse(\%in))
   {
      # Write some info to the logfile:
      # - Date and time
      # - Remotehost
      # - Parameters passed to the program
      
		# get the time and make an extension out of it, e.g. 990622. Notice that
		# position 4 in the return array is the month-1. Position 3 is the day of
		# the month and position 5 is the year. If one of this 3 numbers is smaller
		# than 10 the 0 is put in front of it, i.e. 1->01
      
		@time = localtime( time );
      $time[4]++;
      for( $i = 3; $i < 6; $i++ )
      {
         if( $time[$i] < 10 ){ $time[$i] = "0".$time[$i]; }
      }
      $logfile = $LOG.$time[5].$time[4].$time[3];
      $date = localtime( time );

		# Try to open the log file

      if( !open (LOG, ">> ".$PATH.$logfile ) )
      {
         &PrintError(3, $logfile);
      }
      else
      {
         flock(LOG, 2);
         print LOG $date, "\n";
         print LOG $ENV{'REMOTE_HOST'}, "\n";

         CHECKKEYS: foreach $key ( keys( %in ) )
         {
				# Check for too much data, extremely important!
				if( length($in{$key}) > $LIMIT )
				{
      		   &PrintError(6);
					$toomuchdata = 1;
         	   print LOG $key, " ERROR: TOO MUCH INPUT DATA SUBMITTED!\n";
					last CHECKKEYS;
				}

				# Take care of linebreaks on withdraw
            if( $key eq 'withdraws' )
            {
               # Replace linebreaks with <br>
               $in{$key} =~ tr/\r\f/\n/;
               $in{$key} =~ s/\n\n/\n/g;
               $in{$key} =~ s/\n\n/\n/g;
               $in{$key} =~ s/\n/<br>/g;          
            }
            else
            {
               # remove linebreaks and double spaces
               $in{$key} =~ tr/\r\n\f/ /;
               $in{$key} =~ s/  / /g;
            }
            print LOG $key, " ", $in{$key}, "\n";
         }
         print LOG "\n";
         flock(LOG, 8);
         close(LOG);
         chmod 0600, $PATH.$logfile;

         ACTIONS:
         {
				# Too much data?

				if( $toomuchdata )
				{
					last ACTIONS;
				}

            # Check wether parameters come from the short or long menu
    
            if( exists $in{'action'} )
            {
               # Parameters coming from short menu
               
               SWITCH:
               {
                  # Add a new entry to the database
 
                  if( $in{'action'} eq "add" )
                  {
                     if( !&ReadData( \@data, $DATAFILE ) )
                     {
                        &PrintTable( \@data );
                        &LongMenu;
                     }
                     last SWITCH;
                  }
               
                  # Change an entry of the database. Here it is important that the modification
                  # time of the databasefile is also read. This time is also passed to
                  # &LongMenu, from where this program is again called with a form and the
                  # modification time is also passed to the program. More details further down,
                  # when the element is added.
               
                  if( $in{'action'} eq "change" )
                  {
                     if( !&ReadData( \@data, $DATAFILE ) )
                     {
                        $time = &GetTime( $DATAFILE );
                        
                        # $time contains the actual modification time of the file, $in{'time'}
                        # contains the modification time of the database, when the user last
                        # loaded it and decided to make a change. Obviously in the meantime
                        # somebody else changed the database. So the database file has to be
                        # read and printed again. The ShortMenu has to be presented again 

                        if( $time > $in{'time'} )
                        {
                           &PrintError(0);
                           &PrintTable( \@data );
                           &PrintError(0);
                           &ShortMenu( $data[0], $time );
                        }
                        else
                        {
                           &PrintTable( \@data );
                           &LongMenu( $in{'change_no'}, \@data, $time );
                        }
                     }
                     last SWITCH;
                  }
               
                  # Delete an entry from the database. The modification time of the databasefile
                  # and the backup file are both passed to &ShortMenu.
               
                  if( $in{'action'} eq "delete" )
                  {
                     if( !&ReadData( \@data, $DATAFILE ) )
                     {
                        $time = &GetTime( $DATAFILE );
                        if( $time > $in{'time'} )
                        {
                           &PrintError(0);
                           &PrintTable( \@data );
                           &PrintError(0);
                           &ShortMenu( $data[0], $time );
                        }
                        else
                        {                 
                           if( !&WriteData( \@data, $BACKUP ) )
                           {
                              $baktime = &GetTime( $BACKUP );
                              if( !&DeleteElement( $in{'delete_no'}, \@data, $logfile) )
                              {
                                 if( !&WriteData( \@data, $DATAFILE ) )
                                 {
                                    $time = &GetTime( $DATAFILE );
                                    &PrintTable( \@data );
                                    &ShortMenu( $data[0], $time, $baktime );
                                 }
                              }
                           }
                        }
                     }
                     last SWITCH;
                  }
               
                  # Exchange two entries from the database
               
                  if( $in{'action'} eq "putto" )
                  {
                     if( !&ReadData( \@data, $DATAFILE ) )
                     {
                        $time = &GetTime( $DATAFILE );
                        if( $time > $in{'time'} )
                        {
                           &PrintError(0);
                           &PrintTable( \@data );
                           &PrintError(0);
                           &ShortMenu( $data[0], $time );
                        }
                        else
                        {
                           &PutTo( $in{'putto_source'}, $in{'putto_target'}, \@data );
                           if( !&WriteData( \@data, $DATAFILE ) )
                           {
                              $time = &GetTime( $DATAFILE );
                              &PrintTable( \@data );
                              &ShortMenu( $data[0], $time );
                           }
                        }
                     }
                     last SWITCH;
                 }
               
                  # Undo last delete. The actuall modification times of the database file and
                  # the backup file are compared with the ones right after the delete happened.
                  # If meanwhile somebody else has changed at least one of the two, no undo can
                  # be performed.
               
                  if( $in{'action'} eq "undo" )
                  {
                     $time = &GetTime( $DATAFILE );
                     $baktime = &GetTime( $BACKUP );
                     if( $time > $in{'time'} or $baktime > $in{'baktime'} )
                     {
                        &PrintError(1);
                     }
                     else
                     {
                        if( !&ReadData( \@data, $BACKUP ) )
                        {
                           if( !&WriteData( \@data, $DATAFILE ) )
                           {
                              $time = &GetTime( $DATAFILE );
                              &PrintTable( \@data );
                              &ShortMenu( $data[0], $time );
                           }
                        }
                     }
                     last SWITCH;
                  }
               } # SWITCH
               last ACTIONS;
            } # end check if parameters coming from short menu
         
            # Parameters coming from the long menu contain the parameter no
         
            if( exists $in{'no'} )
            {
               if( !&ReadData( \@data, $DATAFILE ) )
               {
            
                  if( $in{'no'} == 0 )
                  {
                     # Now really add the new element to the data;
               
                     &AddElement( \%in, \@data );
                  }
                  else
                  {
                     # Now really change the element. If the modification time of the database
                     # file has changed in the mean time, the change cannot be performned
               
                     $time = &GetTime( $DATAFILE );
                     if( $time > $in{'time'} )
                     {
                        &PrintError(2);
                        last ACTIONS;
                     }
                     else
                     {
                        &ChangeElement( \%in, \@data );
                     }
                  }
                  if( !&WriteData( \@data, $DATAFILE ) )
                  {
                     $time = &GetTime( $DATAFILE );
                     &PrintTable( \@data );
                     &ShortMenu( $data[0], $time );
                  }
                  last ACTIONS;
               }
            }
         } # ACTIONS
      } # if: check if it's possible to open the log file
   } # if: check if any parameter was passed to the program
   else
   {
      # This is the part which is done, when the program is called without
      # any parameters, which means it is NOT called from ShortMenu or LongMenu.
      # The result is just the tabell and the ShortMenu. Note that the time
      # of the data the user gets is read!
      
      if( !&ReadData( \@data, $DATAFILE ) )
      {
         $time = &GetTime( $DATAFILE );
         &PrintTable( \@data );
         &ShortMenu( $data[0], $time );
      }
   }

   print '<p><a href="', $BACKTOLINK, '">', $BACKTO, '</a></p>';
   print "\n\n";
   print "<p><hr width=", '"100%"', "></p>\n\n";
   print "</center>\n\n";
   print "<address>\n   Samples / <a href=", '"mailto:', $CONTACTMAIL, '"', ">";
   print $CONTACTMAIL, "</a> / revised ";
   print $REVISED, "\n</address>\n\n";
   print &HtmlBot;
}



############################################
# Print some red box with an error message #
# 990622                                   #
############################################

sub PrintError
{
   my ( $no, $filename )= @_;
   
   if( !defined( $filename ) ){ $filename = ""; }
   
   if( $no == 0 )
   {
      print '<p class="redred"><strong>The data had to be reloaded, because it changed!</strong></p>',"\n\n";
   }
   
   if( $no == 1 )
   {
      print '<div class="redred">',"\n   <strong>\n";
      print "      <p>Sorry, but you cannot undo your last delete, because meanwhile\n";
      print "         somebody else has changed the database file. The deleted\n";
      print "         entry is not lost, please contact me, if you want to get it\n";
      print "         back.</p>\n";
      print "      <p><center>", $CONTACT, "</center></p>\n   </strong>";
      print "</div>\n";
   }
   
   if( $no == 2 )
   {
      print '<div class="redred">',"\n   <strong>\n";
      print "      <p>Sorry, but you cannot change the entry, because meanwhile\n";
      print "         somebody else has changed the database file. Your changes\n";
      print "         are not lost, please contact me, if you want your changes\n";
      print "         to be added to the database file.</p>\n";
      print "      <p><center>", $CONTACT, "</center></p>\n   </strong>";
      print "</div>\n";
   }

   if( $no == 3 )
   {
      print '<div class="redred">',"\n   <strong>\n";
      print "      <p>Serious error, can't write the log file";
      if( $filename ne "" ){ print " ", $filename; }
      print " !</p>\n";
      print "      <p><center>Please contact me.</center></p>\n";
      print "      <p><center>", $CONTACT, "</center></p>\n   </strong>";
      print "</div>\n";
	}

   if( $no == 4 )
   {
      print '<div class="redred">',"\n   <strong>\n";
      print "      <p>Serious error, can't write the database file";
      if( $filename ne "" ){ print " ", $filename; }
      print " !</p>\n";
      print "      <p><center>Please contact me.</center></p>\n";
      print "      <p><center>", $CONTACT, "</center></p>\n   </strong>";
      print "</div>\n";
	}

   if( $no == 5 )
   {
      print '<div class="redred">',"\n   <strong>\n";
      print "      <p>Serious error, can't read the database file";
      if( $filename ne "" ){ print " ", $filename; }
      print " !</p>\n";
      print "      <p><center>Please contact me.</center></p>\n";
      print "      <p><center>", $CONTACT, "</center></p>\n   </strong>";
      print "</div>\n";
	}

   if( $no == 6 )
   {
      print '<div class="redred">',"\n   <strong>\n";
      print "      <p>Serious error, you tried to submit too much data.\n";
      print "         Contact me if you really need to submitt that much data.</p>\n";
      print "      <p><center>", $CONTACT, "</center></p>\n   </strong>";
      print "</div>\n";
   }
}



#######################################
# Get the modification time of a file #
#######################################

sub GetTime
{
   my $filename = shift;
   my @status;
   my @time;
   my $result;
   my $i;
   
   $filename = $PATH.$filename;

   # ( $dev, $ino, $mode, $nlink, $uid, $gid, $rdev, $size, $atime,
   #      $mtime, $ctime,$blksize,$blocks) = stat($filename);
   #      ------
   # the modification time is what we want

   @status = stat( $filename );

   # ( $sec, $min, $hour, $mday, $mon, $year, $wday, $yday, $isdst ) = localtime(time)
   #   -------------------------------------
   # That's what we want in reverse order

   @time = reverse( localtime( $status[9] ) );
      
   # Same trick as described above, numbers smaller than 10 in these fields get
   # a leading zero

   for( $i = 3; $i <= 8; $i++ )
   {
      if( $time[$i] < 10 ){ $time[$i] = "0".$time[$i]; }
   }
   $result = join( '', @time[3..8] );   
}



###############################
# Add an element to the table #
# 980130                      #
###############################

sub AddElement
{
   my ( $in, $data ) = @_;

   # Add the data
   
   push( @$data, ( $$in{'keyword'},
                   $$in{'sample'},
                   $$in{'owner'},
                   $$in{'description'}, 
                   $$in{'prepareDate'},
                   $$in{'place'},
                   $$in{'mass'},
                   $$in{'withdraws'},
                   $$in{'signature'} ) );
   $$data[0]++;
}



##################################
# Change an element of the table #
# 980130                         #
##################################

sub ChangeElement
{
   my ( $in, $data ) = @_;
  
   # Put the data at the right position
   
   @$data[($$in{'no'}*$TABLECOLS-$TABLECOLS+1)..$$in{'no'}*$TABLECOLS] = ( $$in{'keyword'},
                                                                           $$in{'sample'},
                                                                           $$in{'owner'},
                                                                           $$in{'description'},
                                                                           $$in{'prepareDate'},
                                                                           $$in{'place'},
                                                                           $$in{'mass'},
                                                                           $$in{'withdraws'},
                                                                           $$in{'signature'} );
}

####################################
# Delete an element from the table #
# 980130                           #
####################################

sub DeleteElement
{
   my ( $no, $data, $logfile ) = @_;
   my $date;
   my $key;
   my $i;
   my $ioerror = 0;
   
   # First write the data, which should be deleted to the logfile
   
   $date = localtime( time );
   if( !open (LOG, ">> ".$PATH.$logfile ) )
   {
       &PrintError(3, $logfile);
       $ioerror = 1;
   }
   else
   {
      flock(LOG, 2);
      print LOG $date, "\n";
      print LOG $ENV{'REMOTE_HOST'}, "\n";
      for( $i = ($no*$TABLECOLS-$TABLECOLS+1); $i <= $no*$TABLECOLS; $i++ )
      {
         print LOG $$data[$i], "\n";
      }
      print LOG "\n";
      flock(LOG, 8);
      close(LOG);
      chmod 0600, $PATH.$logfile;
   
      # Move the data, so that one element disappears
   
      @$data[($no*$TABLECOLS-$TABLECOLS+1)..$$data[0]*$TABLECOLS-$TABLECOLS] = @$data[$no*$TABLECOLS+1..$$data[0]*$TABLECOLS];
      $$data[0]--;
   }
   return $ioerror;
}



####################################
# Put one element to another place #
# 980524                           #
####################################

sub PutTo
{
   my ( $source, $target, $data ) = @_; 
	  
   if( $source > $target )
   {
      @$data[($target-1)*$TABLECOLS+1..$source*$TABLECOLS] = ( @$data[ ($source-1)*$TABLECOLS+1 .. $source*$TABLECOLS     ],
                                                               @$data[ ($target-1)*$TABLECOLS+1 .. ($source-1)*$TABLECOLS ] );
   }
   
   if( $source < $target )
   {
      @$data[($source-1)*$TABLECOLS+1..$target*$TABLECOLS] = ( @$data[     $source*$TABLECOLS+1 .. $target*$TABLECOLS ],
                                                               @$data[ ($source-1)*$TABLECOLS+1 .. $source*$TABLECOLS ] );
   }  
  
}



######################################
# Write the data to the databasefile #
# 990622                             #
######################################

sub WriteData
{
   my ( $data, $filename ) = @_;
   my $i;
   my $j;
   my $line;
   my $ioerror = 0;
      
   if( !open (DATA, "> ".$PATH.$filename ) )
   {
      &PrintError(4, $filename);
      $$data[0] = 0;
      $ioerror = 1;
   }
   else
   {
      flock(DATA, 2);
        
      for( $i = 1; $i <= $$data[0]; $i++ )
      {
         print DATA "/new\n";
         for( $j = 1; $j < ($TABLECOLS+1); $j++ )
         {
            print DATA $$data[($i-1)*$TABLECOLS+$j], "\n";
         }
         print DATA "\n";
      }
      
      flock(DATA, 8);
      close DATA;
		chmod 0640, $PATH.$filename;
   }
   return $ioerror;
}



############################################################################
# Read in the data from the databasefile and store in in the variable data #
# 990622                                                                   #
############################################################################

sub ReadData
{
   my ( $data, $filename ) = @_;
   my $line;
   my $i;
   my $ioerror = 0;
         
   if( -e $PATH.$filename )
   {
      push( @$data, 0 );

      if( !open (DATA, "< ".$PATH.$filename ) )
      {
         &PrintError(5, $filename);
         $$data[0] = 0;
         $ioerror = 1;
      }
      else
      {
         flock(DATA, 2);
         while ($line=<DATA>)
         {
            if( $line =~ /\/new/ )
            {
               $$data[0]++;
               for( $i = 0; $i < $TABLECOLS; $i++ )
               {
                  $line=<DATA>;
                  chop( $line );
                  push( @$data, $line );
               }
            }
         }
         flock(DATA, 8);
         close DATA;
      }
   }
   else
   {
      $$data[0] = 0;
   }
   return $ioerror;
}



#####################################
# Print the data in form of a table #
# 980130                            #
#####################################

sub PrintTable
{
   my $data = shift;
   my $i;
   my $j;
   my $text;

   # Print the headlines of the table elements

   print '<p><table border="0" cellspacing="2" cellpadding="2"', ">\n";
   print '   <tr><th class="samples">No</th>', "\n";
   print '       <th class="samples">Keyword</th>', "\n";
   print '       <th class="samples">Sample</th>', "\n";
   print '       <th class="samples">Owner</th>', "\n";
   print '       <th class="samples">Description/Preparation</th>', "\n";
   print '       <th class="samples">Date of Preparation or Delivery</th>', "\n";
   print '       <th class="samples">Place</th>', "\n";
   print '       <th class="samples">Original Mass</th>', "\n";
   print '       <th class="samples">Withdraws (Mass,Date,Person)</th>', "\n";
   print '       <th class="samples">Date, Signature</th>', "\n";
	print "   </tr>\n";

   for( $i = 1; $i <= $$data[0]; $i++ )
   {
      print '   <tr align=center><td class="samples">', $i, "</td>\n";
      for( $j = 1; $j < ($TABLECOLS+1); $j++ )
      {
         $text = $$data[($i-1)*$TABLECOLS+$j];
         
         # This makes sure that an table element has borders, even though the
         # element is empty
         
         if( $text eq '' ){ $text = '&nbsp' }
         print '       <td class="samples">', $text, "</td>\n";
      }
      print "   </tr>\n";
   }
   print "</table></p>\n\n";
}



###########################################################################
# Create the input form for the the actions, that can be done on the data #
# 980130                                                                  #
###########################################################################

sub ShortMenu
{
   my ( $no, $time, $baktime )= @_;
   my $i;
   
   if( !defined( $baktime ) ){ $baktime = 0 }
   if( !defined( $time ) ){ $time = 0 }

   print "<p><hr width=", '"100%"', "></p>\n\n";
   print '<p><form method="post" action="', $ACTION, '#table">',"\n";
   print '   <table cellspacing=5>',"\n      ";
   print '<tr><td';
   
   print ' rowspan=', 3 * ( $no > 0 ) + ( $baktime > 0 and $time > 0 ) + 1, ' valign=middle';
   
   print '><input type=submit value="Do This:"></td>';
   print "\n      ";
   print '    <td><input type="radio" name="action" value="add" checked></td>';
   print "\n      ";
   print '    <td>Add New Entry</td>',"\n      ";
   
   if( $no > 0 )
   {    
      print '<tr><td><input type="radio" name="action" value="change"></td>',"\n      ";
      print '    <td>Change Entry</td>',"\n      ";
      print '    <td>No.</td>',"\n      ";
      print '    <td><select name="change_no">',"\n      ";
      for( $i=1; $i <= $no; $i++ )
      {
         print '           <option>', $i, "\n      "; 
      }
      print '        </select></td>', "\n      ";
      print '<tr><td><input type="radio" name="action" value="delete"></td>',"\n      ";
      print '    <td>Delete Entry</td>',"\n      ";
      print '    <td>No.</td>',"\n      ";
      print '    <td><select name="delete_no">',"\n      ";
      for( $i=1; $i <= $no; $i++ )
      {
         print '           <option>', $i, "\n      "; 
      }
      print '        </select></td>', "\n   ";
      print '<tr><td><input type="radio" name="action" value="putto"></td>',"\n      ";
      print '    <td>Entry</td>',"\n      ";
      print '    <td>No.</td>',"\n      ";
      print '    <td><select name="putto_source">',"\n      ";
      for( $i=1; $i <= $no; $i++ )
      {
         print '           <option>', $i, "\n      "; 
      }
      print '        </select></td>', "\n      ";
      print '    <td>Should Become Entry</td>',"\n      ";
      print '    <td>No.</td>',"\n      ";
      print '    <td><select name="putto_target">',"\n      ";
      for( $i=1; $i <= $no; $i++ )
      {
         print '           <option>', $i, "\n      "; 
      }
      print '        </select></td>', "\n      ";
   }  
   
   # If the ShortMenu is called with both $time and $baktime, it means that before that was a Delete,
   # which can be undone, so the menu gets the Undo option. If ShortMenu is only called with time
   # This time is just passed on to the next call of the program
   
   if( $baktime > 0 )
   {
      print '   <tr><td><input type="radio" name="action" value="undo"></td>',"\n      ";
      print '    <td>Undo Last Delete</td>',"\n";
      print "   </table>\n";
      print '   <input type=hidden name="baktime" value=', $baktime, '>', "\n";
   }
   else
   {
      print "   </table>\n";
   }
   print '   <input type=hidden name="time" value=', $time, '>', "\n";
   print "</form></p>\n\n";
}


#############################################
# Create the input form to enter the table  #
# 980130                                    #
#############################################

sub LongMenu
{
   my ( $no, $data, $time ) = @_;
   my @options;
   my $v1;
   my $v2; 
   
   if( $#$data == -1 )
   {
      # @options should have $TABLECOLS elements!
      
      @options = ('','','','','','','','','');
      $no = 0;
      $v1 = '';
      $v2 = '';
      $time = 0;
   }
   else
   {
      @options = @$data[($no*$TABLECOLS-$TABLECOLS+1)..$no*$TABLECOLS];
      $v1 = ' value="';
      $v2 = '"'; 
      &LineBreaks( \$options[3] );
      $options[7] =~ s/<br>/\n/g;
   }
   
   print '<a name="table"></a>', "\n\n";
   print "<p><hr width=", '"100%"', ">\n\n";
   print '<div class="redblack">',"\n";
   print "<table>\n   <tr><td><tt>&lt;sup&gt;...&lt;/sup&gt;</tt></td>\n";
   print "       <td>This makes a superscript like in <sup>15</sup>N.</td>\n";
   print "   <tr><td><tt>&lt;sub&gt;...&lt;/sub&gt;</tt></td>\n";
   print "       <td>This makes a subscript like in C<sub>2</sub>.</td>\n";
   print "   <tr><td><tt>&lt;br&gt;</tt></td>\n";
   print "       <td>Normally linebreaks are done automatically, use this to make an";
   print " explicit linebreak.</td>\n</table>\n</div>\n\n";

   print '<form method="post" action="', $ACTION, '">',"\n";
   print '   <table valign="middle">',"\n      ";

   print '<tr><td>Keyword:</td>',"\n      ";
   print '    <td><input type="text" name="keyword"', $v1, $options[0], $v2;
   print                ' size=', $INPUTCOLS, ' maxlength=', $INPUTCOLS, '></td>', "\n      ";

   print '<tr><td>Sample:</td>',"\n      ";
   print '    <td><input type="text" name="sample"', $v1, $options[1], $v2;
   print                ' size=', $INPUTCOLS, ' maxlength=', 2*$INPUTCOLS, '></td>', "\n      ";

   print '<tr><td>Owner:</td>',"\n      ";
   print '    <td><input type="text" name="owner"', $v1, $options[2], $v2;
   print                ' size=', $INPUTCOLS, ' maxlength=', $INPUTCOLS, '></td>', "\n      ";

   print '<tr><td>Description/Preparation:</td>',"\n      ";
   print '    <td><textarea name="description" rows=5 cols=', $INPUTCOLS, '>';
   print                    $options[3];        
   print         '</textarea></td>', "\n      ";

   print '<tr><td>Date of Preparation or Delivery:</td>',"\n      ";
   print '    <td><input type="text" name="prepareDate"', $v1, $options[4], $v2;
   print                ' size=', $INPUTCOLS, ' maxlength=', $INPUTCOLS, '></td>', "\n      ";

   print '<tr><td>Place:</td>',"\n      ";
   print '    <td><input type="text" name="place"', $v1, $options[5], $v2;
   print                ' size=', $INPUTCOLS, ' maxlength=', $INPUTCOLS, '></td>', "\n      ";

   print '<tr><td>Original Mass:</td>',"\n      ";
   print '    <td><input type="text" name="mass"', $v1, $options[6], $v2;
   print                ' size=', $INPUTCOLS, ' maxlength=', $INPUTCOLS, '></td>', "\n      ";

   print '<tr><td>Withdraws<br>(Mass, Date, Person):</td>',"\n      ";
   print '    <td><textarea name="withdraws" rows=5 cols=', $INPUTCOLS, '>';
   print                    $options[7];        
   print         '</textarea></td>', "\n      ";

   print '<tr><td>Date, Signature:</td>',"\n      ";
   print '    <td><input type=text name="signature"', $v1, $options[8], $v2;
   print                ' size=', $INPUTCOLS, ' maxlength=', $INPUTCOLS, '></td>', "\n      ";

   print '<tr><td colspan=2 align=center><input type=reset value="Reset Fields">';
   print                                '<input type=submit value="OK"></td>', "\n";
   print "   </table>\n   ";
   print '<input type=hidden name="no" value=', $no, '>', "\n";
   if( $time > 0 )
   {
      print '   <input type=hidden name="time" value=', $time, '>', "\n";
   }
   print "</form>","\n\n";
}



#######################################################################
# Insert linebreaks instead of spaces to make the text easier to read #
#######################################################################

sub LineBreaks
{
   my $str = shift;
   my @pos=(0);
   my @chars;
   my $i;
   my $break = $INPUTCOLS-1;
   
   @chars = split( //, $$str );
   push( @chars, ' ' );
  
   for( $i = 0; $i <= $#chars; $i++ )
   {
      if( $chars[$i] eq ' ' )
      {
         push( @pos, $i );
         if( $pos[1] > $break )
         {
            if( $pos[1] == $break+1 )
               { $chars[$pos[1]] = "\n"; }
            else
               { $chars[$pos[0]] = "\n"; }
            $break = $pos[0] + $INPUTCOLS;
         }
         shift( @pos );
      }
   }
   $chars[$#chars] = '';
   $$str = join( '', @chars );
}
